/*
 * act0.cpp
 * Contains definitions for checking action 0s.
 *
 * Copyright 2005-2006 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<string>
#ifdef _MSC_VER
#pragma warning(disable:4702)//unreachable code
#include<vector>
#pragma warning(default:4702)
#else
#include<vector>
#endif
#include<cassert>
#include<errno.h>

using namespace std;

#include"renum.h"
#include"inlines.h"
#include"messages.h"
#include"ExpandingArray.h"
#include"sanity_defines.h"
#include"data.h"

/*Data Format:
chars 01,02,04 indicate bytes,words,doubles
char 03 extended byte
char FE means variable length, see corresponding LengthData struct in subData
char FF means no such property

for variable length properties only:
char r means the next char is repeated some number of times, set by the char
  after it, for which chars 80-FF take on the value of the data indicated by
  char num&0x7F of this string (well, sorta--FEs aren't counted, for one)
char x means multiply the next two chars together
char * means the next char is repeated an arbitrary number of times. This is
  followed by the length of the terminator (1, 2, or 4) and then the
  terminator
char l means the next character appears in the NFO
char | means what is on either side is valid -- the left side MUST have at
least one literal. (ex: \x03 == l\xFF\x02|\x01)
char FE means variable, see corresponding LengthData struct in subData
A backslash is used to escape nulls that are not end-of-string characters
all other characters are undef
*/

typedef basic_string<uchar> ustring;
typedef vector<int> int_str;

class PropData:public GlobalPtr<PropData>{
public:
    PropData(){}
    void Init(FILE*);
    uchar GetData(uint)const;
    uint GetLength()const{return(uint)length.length();}
    const PropData*GetVarLength(int)const;
    uint GetRepeat(uint&,const int_str&)const;
private:
    ustring length;
    static ustring readString(FILE*);
    static int CountFE(const ustring&);
    void operator=(const PropData&);
    PropData(const PropData&);
};

class Check0:public GlobalPtr<PropData>{
public:
    void Check(const string&);
    static Check0&Instance(){static Check0 obj;return obj;}
    uint GetFeat8(){return _p[8].GetLength();}
private:
    bool Load();
    Check0(){Load();}
    bool CheckVar(uint&,const string&,const PropData&);
    void operator=(const Check0&);
    Check0(const Check0&);
    bool fail;
};

struct Feat8sub{
    uint maxfirst;
    uint maxlast;
};

class Feat8:public GlobalPtr<Feat8sub>{
public:
    SINGLETON(Feat8);
};

Feat8::Feat8(){
    uint numprops=Check0::Instance().GetFeat8();
    _p=new Feat8sub[numprops];
    FILE*pFile=myfopen(dat0f8);
    for(uint i=8;i<numprops;i++){
        _p[i].maxfirst=GetCheckByte(dat0f8);
        _p[i].maxlast=GetCheckByte(dat0f8);
    }
    fclose(pFile);
}

void Check0(const string&str){
    Check0::Instance().Check(str);
}

uchar PropData::GetData(unsigned int prop)const{
	return prop<length.length()?length[prop]:(uchar)0xFF;
}

const PropData*PropData::GetVarLength(int prop)const{
    PropData*ret=_p;
    for(int i=0;i<prop;i++)
        if(length[i]==0xFE)ret++;
    return ret;
}

uint PropData::GetRepeat(uint&len_off,const int_str&decoded)const{
    uint repeat=GetData(len_off);
    if(repeat=='x'){
        repeat=GetRepeat(++len_off,decoded);
        return repeat*GetRepeat(++len_off,decoded);
    }
    if(repeat=='l')
        return GetData(++len_off);
    if(repeat&0x80)
        return decoded[repeat&0x7F];
    return repeat;
}

void PropData::Init(FILE*pFile){
    int j=CountFE(length=readString(pFile));
    if(j){
        _p=new PropData[j];
        for(int k=0;k<j;k++)
            _p[k].Init(pFile);
    }
}

ustring PropData::readString(FILE*pFile){
    ustring ret;
    int ch;
    bool escape=false;
    do{
        if((ch=fgetc(pFile))==EOF)
            throw 0;
        if(!escape){
            if(!ch)break;
            if(ch=='\\'){
                escape=true;
                continue;
            }
        }
        escape=false;
        ret.append(1,(char)ch);
    }while(true);
    return ret;
}

int PropData::CountFE(const ustring&str){
    int ret=0;
    size_t start=0;
    while((start=str.find_first_of((unsigned char)0xFE,start)+1)!=0)ret++;
    return ret;
}

void Check0::Check(const string&str){
    if(fail)return;
    assert(extract_byte(str,0)==0);
    int feature=extract_byte(str,1);
    if(!IsValidFeature(ACT0,feature)){
        IssueMessage(FATAL,INVALID_FEATURE);
        return;
    }
    unsigned int i=4+extended_len(str,4),firstID=extract_extended(str,4);
    int propsRemain=extract_byte(str,2),IDs=extract_byte(str,3),prop=-1,len,j;
    uint maxID=firstID+IDs-1;
    if(propsRemain==0)
        IssueMessage(WARNING1,NO_PROPS);
    if(IDs==0&&(propsRemain!=1||feature||extract_byte(str,i)!=0x1A))IssueMessage(WARNING1,NO_IDS);
    if(feature!=8&&IDs)CheckID(feature,firstID)&&CheckID(feature,maxID);
    Expanding0Array<uint>propLoc;
    try{
        while(propsRemain){
            try{
                prop=extract_byte(str,i);
                if(feature==8){
                    if(firstID>Feat8::Instance()[prop].maxfirst)
                        IssueMessage(ERROR,INVALID_ID,firstID,0,Feat8::Instance()[prop].maxfirst);
                    if(maxID>Feat8::Instance()[prop].maxlast)
                        IssueMessage(ERROR,INVALID_ID,maxID,0,Feat8::Instance()[prop].maxlast);
                }
            }catch(unsigned int){
                IssueMessage(ERROR,INSUFFICIENT_PROPS,propsRemain);
                return;
            }
            len=_p[feature].GetData(prop);
            if(len==0xFF){
                IssueMessage(FATAL,INVALID_PROP,i,prop);
                return;
            }
            if(propLoc[prop])
                IssueMessage(WARNING2,REPEATED_PROP,i,prop,propLoc[prop]);
            propLoc[prop]=i;
            if(len==0xFE){
                i++;
                const PropData*data=_p[feature].GetVarLength(prop);
                for(j=0;j<IDs;j++)
                    if(!CheckVar(i,str,*data))return;
            }else if(len==3){
                i++;
                for(j=0;j<IDs;j++)
					i+=extended_len(str,i);
            }else
                i+=len*IDs+1;
            propsRemain--;
        }
        if(2*i<str.length())
            IssueMessage(WARNING2,EXTRA_DATA,i);
        if(2*i>str.length())
            IssueMessage(ERROR,INSUFFICIENT_DATA2,i-str.length()/2,prop);
    }catch(uint off){
        IssueMessage(FATAL,INSUFFICIENT_DATA,prop,EXPECTED_BYTES(off),EXPECTED_LOC(off));
    }
}

bool Check0::CheckVar(uint&str_loc,const string&str,const PropData&vdata){
    bool findPipe=false;
    uchar ch;
    uint orig_loc=str_loc,vdatalen=vdata.GetLength();
    int char_data;
    int_str decoded;
    for(uint i=0;i<vdatalen;i++){
        ch=vdata.GetData(i);
        if(findPipe&&ch!='|')continue;
        else if(findPipe){
            findPipe=false;
            continue;
        }
        switch(ch){
        case'|':return true;
        case'l':
            if(vdata.GetData(++i)!=extract_byte(str,str_loc)){
                str_loc=orig_loc;
                findPipe=true;
            }else
                str_loc++;
            break;
        case'r':{
            const PropData*vdata2=vdata.GetVarLength(i);
            uchar repeat_data=vdata.GetData(++i);
            uint times=vdata.GetRepeat(++i,decoded);
            if(repeat_data<5)
                str_loc+=repeat_data*times;
            else if(repeat_data==0xFE){
                for(uint j=0;j<times;j++)
                    if(!CheckVar(str_loc,str,*vdata2))return false;
            }else{
                IssueMessage(0,INVALID_DATAFILE,"0.dat",DAT2,'r',repeat_data);
                fail=true;
                return true;
            }
            break;
        }case 3:
            char_data=extract_byte(str,str_loc++);
            if(char_data==0xFF){
                char_data=extract_word(str,str_loc);
                str_loc+=2;
            }
            decoded.push_back(char_data);
            break;
        case 1:
        case 2:
        case 4:
            decoded.push_back(extract_variable(str,str_loc,ch));
            str_loc+=ch;
            break;
        case'*':{
            const PropData*vdata2=vdata.GetVarLength(i);
            uchar repeat_data=vdata.GetData(++i);
            int term_len=vdata.GetData(++i);
            uint term;
            uint(*extract_term)(const string&,uint);
            switch(term_len){
            case 1:
                term=vdata.GetData(++i);
                extract_term=extract_byte;
                break;
            case 2:
                term=vdata.GetData(++i);
                term|=vdata.GetData(++i)<<8;
                extract_term=extract_word;
                break;
            case 4:
                term=vdata.GetData(++i);
                term|=vdata.GetData(++i)<<8;
                term|=vdata.GetData(++i)<<16;
                term|=vdata.GetData(++i)<<24;
                extract_term=extract_dword;
                break;
            default:
                IssueMessage(0,INVALID_DATAFILE,"0.dat",DAT3,'*',repeat_data,term_len);
                fail=true;
                return false;
            }
            try{
                if(repeat_data<5){
                    while(extract_term(str,str_loc)!=term)
                        str_loc+=repeat_data;
                }else if(repeat_data==0xFE){
                    while(extract_term(str,str_loc)!=term)
                        if(!CheckVar(str_loc,str,*vdata2))return false;
                }else{
                    IssueMessage(0,INVALID_DATAFILE,DAT2,"0.dat",'*',repeat_data);
                    fail=true;
                    return true;
                }
                str_loc+=term_len;
                break;
            }catch(uint){
                IssueMessage(ERROR,MISSING_TERMINATOR);
                return false;
            }
            break;
        }case 0xFE:
            CheckVar(str_loc,str,*vdata.GetVarLength(i));
            break;
        DEFAULT(Check0::CheckVar,ch)
        }
    }
    if(findPipe){
        IssueMessage(ERROR,MISSING_TERMINATOR);
        return false;
    }
    return true;
}

bool Check0::Load(){
    fail=false;
    FILE*pFile=myfopen(dat0);
    _p=new PropData[MaxFeature()+1];
    for(uint i=0;i<=MaxFeature();i++)
        _p[i].Init(pFile);
    fclose(pFile);
    return true;
}
