/*
 * data.cpp
 * Data file contents and implementation of helper functions.
 *
 * Copyright 2005-2006 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef _WIN32
#include"win32.h"
#endif

#include<cstdio>
#include<cstdlib>
#include<cassert>
#include<string>
#include<sys/stat.h>
#include<cstdarg>
#include<errno.h>
#ifndef _WIN32
#   include<unistd.h>
#endif

using namespace std;

#include"renum.h"
#include"globals.h"
#include"data.h"
#include"messages.h"
#include"win32.h"


typedef unsigned int uint;

//Let's dump the data files into the source code. That sounds like fun, right?

static const char _datfeat[]="\x0C\x3F\x1F\x1F\x3F\x1F\x06\x09\x0F\x01\x07\x0D\x0F\x01";

static const char _dat0[]=
"\x02\xFF\x01\x01\x01\x01\x01\x01\x01\x02\xFF\x02\xFF\x01\x04\xFF\xFF\xFF\x01\x01\x01\x01\x01\x01\x01\x01\x01\x02\x01\x04\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00"
"\x02\xFF\x01\x01\x01\xFF\x01\x01\x01\x01\x04\xFF\xFF\xFF\x01\x01\x01\x01\x01\x01\x01\x01\x04\x01\x01\x01\x01\x01\x01\x02\x02\x00"
//                                 08                              10                              18                              20                              28
"\x02\xFF\x01\x01\x01\xFF\x01\x01\x01\x01\x01\x01\x01\x02\xFF\x01\x01\x04\x01\x01\x01\x01\x01\x01\x02\x02\x00"
"\x02\xFF\x01\x01\x01\xFF\x01\x01\x01\x01\x01\x01\x01\x01\x01\x02\xFF\x01\x01\x04\x01\x01\x01\x01\x02\x02\x00"
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x04\xFE\x01\x01\x01\x01\xFE\x01\x02\x01\x04\x01\x01\x01\x00"
  "\x03r\xFE\x80\x00"//                ^^
    "l\\\x00l\\\x00l\\\x00l\\\x00|\x04*\xFE\x01\x80\x00"
      "\x01\x01\x01\x01\x01\x01\x04\x00"
  "*\xFE\x02\\\x00\\\x00\x00"//                            ^^
    "\x01\x01r\x01x\x80\x81\x00"
"\x00"
//                                 08                              10
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x01\x01\x01\x02\xFE\x01\x00"
  "\x01\x01r\x04x\x81\x20\x00"//                       ^^
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x01\x02\x01\x01\x01\x01\x01\x02\x01\x02\x02\x01\x01\x01\x04\x01\x01\x01\x01\x01\x01\x04\x00"
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x04\x02\x04\x02\x04\x04\x02\x00"
//                                 08                              10                              18                              20
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x01\x02\x02\x02\x01\x01\x02\x01\x01\x00"
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x01\xFE\x01\x02\x02\x02\x01\x02\x04\x01\x01\x01\xFE\xFE\x01\x01\x01\x04\x02\x04\x04\x04\x02\x04\x01\x00"
  "\x01\x04r\xFE\x80\x00"//                ^^
    "l\xFE\x01\x01|*\xFE\x02\\\x00\x80\x00"
      "\x00"
      "\x01\x01\xFE\x00"
        "l\xFF|l\xFE\x02|\x01\x00"
          "\x00"//^^
  "\x01r\x01\x80\x00"//                                                                ^^
  "r\x01\x03\x00"//                                                                        ^^
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x02\x02\x02\x02\x02\x02\x01\x01\x01\x04\x01\x01\x01\x02\x04\x00"
//                                 08                              10                              18
"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x01\x01\x00";

static const char _dat2v[]=
    "\x00\x82\x01\x81\x02\x81\x03\x81\x09\x82\x0A\x82\x0C\x81\x10\x84\x18\x84\x1A\x84\x1B\x81\x1C\x84\xFF\xF0"
"\x00\x40\x83\x41\x83\x42\x84\x43\x84\x45\x83\x46\x84\x47\x84\x5F\x81\x60\x81\x73\x7F\x84\x7F\xFF\xF0"
"\x01\x42\x81\x43\x84\x47\x84\x5F\x81\x7F\x84\x7F\xFF\xF0"
"\x02\x42\x81\x43\x84\x47\x84\x5F\x81\x7F\x84\x7F\xFF\xF0"
"\x03\x40\x83\x42\x81\x43\x84\x44\x82\x47\x84\x5F\x81\x7F\x84\x7F\xFF\xF0"
"\x08\x11\x81\x40\x84\x41\x84\x42\x82\x43\x83\x44\x81\x45\x82\x46\x84\x47\x84\x48\x85\x49\x84\x5F\x83\x60\x82\x1F\x61\x81\x1F\x62\x84\x1F\x63\x81\x1F\x64\x82\x1F\x7F\x84\x7F\xFF\xF0"
"\x05\x7F\x84\x7F\x82\x80"
"\x06\x80\x80"
"\x08\x40\x81\x41\x81\x42\x81\x43\x81\x44\x84\x45\x81\x46\x81\x5F\x81\x60\x82\x6B\x61\x82\xFF\x7F\x84\x7F\x80\x80"
"\x08\x40\x81\x41\x81\x7F\x84\x7F\xDE\x80"
"\x0A\x40\x81\x41\x81\x42\x81\x43\x83\x44\x81\x5F\x81\x60\x83\xFF\x7F\x84\x7F\x80\x80"
"\x08\x40\x82\x41\x82\x42\x82\x5F\x82\x7F\x84\x7F\xB6\x80"
"\x0B\x7F\x84\x7F\x80\x80"
"\x0C\x80\x80";
//Format: Low nibble: width (1/2/4) high nibble: bitmask: Read in 7, read in D, write in D, reserved
//                                  80                              88                              90                              98                              A0
static const char _dat79Dv[]="\x20\x00\x00\x00\xC1\xC1\x80\x80\x00\x84\x00\x00\xC4\x00\xC1\xE1\xE4\x00\x00\xC1\xE2\xE2\xE2\xE2\xE1\x00\x24\x00\x00\x00\xC4\xE4\x24";
static const char _datD[]="\x10\x0C\x74\x00\x58\x00\x0B\x00\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1E\x13\x00\x00\x25\x00\x40\x00\x00\x00";
static const char _datIDs[]="\x73\x00\x57\x00\x0A\x00\x28\x00\xFF\x00\x04\x00\x0A\x00\xFF\x00\x00\x00\xFF\x00\x24\x00\x1F\x00\xFF\xFF";
static const char _dat4[]=
"\x04\x04\x04\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00"
"\x1C\x1C\x1C\x1C\x04\x04\x00\x04\x00\x00\x15\x06\x00";
static const char _datB[]="\x03\x04\x02\x02\x02\x03";
static const char _dat5[]="\x0B\x30\x00\x4A\x00\x49\x00\x41\x00\x06\x00\x00\x01\x4C\x00";
//Count of IDs in class:           -0000-                          -2000-                          -4000-                          -6000-                          -8000-                          -A000-                          -C000-                          -E000-
static const char _datTextIDs[]="\x35\x03\x11\x00\x25\x00\x19\x00\x5D\x00\x11\x00\x6D\x00\x08\x00\x11\x00\x3C\x00\x2A\x00\x08\x00\x19\x00\x39\x00\x80\x00\x00\x00\x08\x01\x6C\x00\x38\x00\x43\x00\x44\x00\x00\x00\x06\x00\x00\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x4D\x00\x00\x00\x00\x00\xFB\x00"
    "\x03\xC4\xC5\xC9";
//                             --08--                          --0C--                          --10--
static const char _dat0f8[]="\x30\x30\x00\xFC\x12\x12\x12\x12\x12\x12\x12\x12\x12\x12\x12\x12";

struct dat{
    char*data,*name;
    uint len;
    int ver;
};

#undef _RENUM_DATA_H_INCLUDED_
#undef DATA
#undef DATA_FILE
#define DATA() static const dat data[]={
#define DATA_FILE(name,ver)\
    {(char*)_dat##name,".renum/" #name ".dat",sizeof(_dat##name)-1,ver},\

#include "data.h"

void fail(char*reason,char*cwd,int err){
    IssueMessage(0,CREATE_FAILED,reason,cwd,err);
    free(cwd);
    sleep(5);
    exit(EDATA);
}

FILE*myfopen(files file){
    FILE*pFile=fopen(data[file].name,"rb");
    if(pFile){
        if(fgetc(pFile)==data[file].ver)return pFile;
        fclose(pFile);
    }else{
        struct stat Stat;
        char*cwd=getcwd(NULL,0);
        if(stat(".renum",&Stat)){
            if(errno==ENOENT){
                if(mkdir(".renum",0755))
                    fail(UNKNOWN_FAIL,cwd,errno);
                else{
                    IssueMessage(0,CREATED_DIR,cwd);
                    sleep(5);
                }
            }else fail(UNKNOWN_FAIL,cwd,errno);
        }else if(Stat.st_mode&S_IFREG)fail(IS_FILE,cwd,0);
        free(cwd);
    }
    pFile=fopen(data[file].name,"wb");
    if(!pFile){
        IssueMessage(0,DATAFILE_ERROR,OPEN,data[file].name+6,"(%d)",errno);
        assert(false);
        exit(2);
    }
    fputc(data[file].ver,pFile);
    if(fwrite(data[file].data,1,data[file].len,pFile)!=data[file].len){
        IssueMessage(0,DATAFILE_ERROR,WRITE,data[file].name+6,"");
        assert(false);
        exit(EDATA);
    }
    fclose(pFile);
    pFile=fopen(data[file].name,"rb");
    if(pFile){fgetc(pFile);return pFile;}
    IssueMessage(0,DATAFILE_ERROR,OPEN,data[file].name+6,"(%d)",errno);
    assert(false);
    exit(EDATA);
}

int _CheckEOF(int dat,files file,char*src,int line){
    if(dat==EOF){
        IssueMessage(0,DATAFILE_ERROR,LOAD,data[file].name+6,"(%s:%d)",src,line);
        assert(false);
        exit(EDATA);
    }
    return dat;
}

int _GetCheckWord(FILE*pFile,files file,char*src,int line){
    int ret=fgetc(pFile);
    return ret|(_CheckEOF(fgetc(pFile),file,src,line)<<8);
}
